﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Windows.Threading;
using Microsoft.Phone.Controls;
using Petzold.MusicSynthesis;

namespace Harp
{
    public partial class MainPage : PhoneApplicationPage
    {
        const int SAMPLE_RATE = 22050;

        static readonly Pitch MIN_PITCH = new Pitch(Note.C, 4);
        static readonly Pitch MAX_PITCH = new Pitch(Note.C, 6);
        static readonly int NUM_PITCHES = MAX_PITCH.MidiNumber - MIN_PITCH.MidiNumber + 1;

        StringInfo[] stringInfos = new StringInfo[NUM_PITCHES];
        double xDelta;
        Mixer mixer;
        Dictionary<int, Point> touchDictionary = new Dictionary<int, Point>();

        public MainPage()
        {
            InitializeComponent();

            mixer = new Mixer();

            for (int index = 0; index < NUM_PITCHES; index++)
            {
                Pitch pitch = new Pitch(MIN_PITCH.MidiNumber + index);
                stringInfos[index] = new StringInfo(pitch);

                SawtoothOscillator oscillator = new SawtoothOscillator(SAMPLE_RATE)
                {
                    Frequency = pitch.Frequency
                };

                AttackDecayEnvelope envelope = new AttackDecayEnvelope(SAMPLE_RATE)
                {
                    Input = oscillator,
                    AttackTime = TimeSpan.FromSeconds(0.01),
                    DecayTime = TimeSpan.FromSeconds(5)
                };

                stringInfos[index].Envelope = envelope;

                IntegerMixerInput mixerInput = new IntegerMixerInput(envelope)
                {
                    Space = -1 + 2.0 * index / (NUM_PITCHES - 1)
                };

                mixer.Inputs.Add(mixerInput);
            }

            // Install event handlers
            Loaded += OnLoaded;
            Touch.FrameReported += OnTouchFrameReported;

            DispatcherTimer tmr = new DispatcherTimer { Interval = TimeSpan.FromSeconds(0.1) };
            tmr.Tick += OnTimerTick;
            tmr.Start();
        }

        void OnLoaded(object sender, EventArgs args)
        {
            int count = NUM_PITCHES - 1;
            xDelta = (int)((ContentPanel.ActualWidth - 4) / count);
            double xStart = (int)((ContentPanel.ActualWidth - count * xDelta) / 2) + 0.5;
            Brush cBrush = new SolidColorBrush(Colors.Red);                     // C
            Brush fBrush = new SolidColorBrush(Colors.Blue);                    // F
            Brush nBrush = this.Resources["PhoneForegroundBrush"] as Brush;     // other naturals
            Brush aBrush = new SolidColorBrush(Colors.Gray);                    // accidentals
            double x = xStart;
            double height = ContentPanel.ActualHeight;
            
            foreach (StringInfo stringInfo in stringInfos)
            {
                Pitch pitch = stringInfo.Pitch;
                double yTop = pitch.IsNatural ? height / 3 : 0;
                double yEnd = pitch.IsNatural ? height : 2 * height / 3;

                stringInfo.TopPoint = new Point(x, yTop);
                stringInfo.EndPoint = new Point(x, yEnd);

                PathFigure pathFigure = new PathFigure
                {
                    StartPoint = stringInfo.TopPoint,
                    IsClosed = false,
                    IsFilled = false
                };

                stringInfo.BezierSegment = new QuadraticBezierSegment
                {
                    Point1 = new Point(x, (yTop + yEnd) / 2),
                    Point2 = stringInfo.EndPoint
                };

                pathFigure.Segments.Add(stringInfo.BezierSegment);
                PathGeometry pathGeometry = new PathGeometry();
                pathGeometry.Figures.Add(pathFigure);
                Path path = new Path
                {
                    Data = pathGeometry,
                    Stroke = pitch.Note == Note.C ? cBrush : (pitch.Note == Note.F ? fBrush : (pitch.IsNatural ? nBrush : aBrush))
                };

                ContentPanel.Children.Add(path);
                x += xDelta;
            }
        }

        protected override void OnNavigatedTo(NavigationEventArgs args)
        {
            DynamicPcmStreamSource dynamicPcmStreamSource = new DynamicPcmStreamSource(SAMPLE_RATE, 15);
            dynamicPcmStreamSource.SampleProvider = mixer;
            mediaElement.SetSource(dynamicPcmStreamSource);

            base.OnNavigatedTo(args);
        }

        void OnTouchFrameReported(object sender, TouchFrameEventArgs args)
        {
            TouchPointCollection touchPoints = args.GetTouchPoints(ContentPanel);

            foreach (TouchPoint touchPoint in touchPoints)
            {
                Point pt = touchPoint.Position;
                int id = touchPoint.TouchDevice.Id;

                switch (touchPoint.Action)
                {
                    case TouchAction.Down:
                        foreach (StringInfo stringInfo in stringInfos)
                        {
                            if (pt.Y >= stringInfo.TopPoint.Y &&
                                pt.Y <= stringInfo.EndPoint.Y &&
                                Math.Abs(pt.X - stringInfo.TopPoint.X) < xDelta / 2)
                            {
                                stringInfo.Envelope.Attack();
                                break;
                            }
                        }
                        touchDictionary.Add(id, pt);
                        break;

                    case TouchAction.Move:
                        Point ptLast = touchDictionary[id];

                        foreach (StringInfo stringInfo in stringInfos)
                        {
                            Point ptTop = stringInfo.TopPoint;
                            Point ptEnd = stringInfo.EndPoint;

                            // Corresponds with string on vertical axis
                            if ((pt.Y >= ptTop.Y || ptLast.Y >= ptTop.Y) &&
                                (pt.Y <= ptEnd.Y || ptLast.Y <= ptEnd.Y))
                            {
                                if ((pt.X > ptLast.X && pt.X >= ptTop.X && ptTop.X >= ptLast.X) ||
                                    (pt.X < ptLast.X && pt.X <= ptTop.X && ptTop.X <= ptLast.X))
                                {
                                    stringInfo.Envelope.Attack();
                                }
                            }
                        }
                        touchDictionary[id] = pt;
                        break;

                    case TouchAction.Up:
                        touchDictionary.Remove(id);
                        break;
                }
            }
        }

        void OnTimerTick(object sender, EventArgs args)
        {
            // Wait until bezier segments are created
            if (stringInfos[0].BezierSegment == null)
                return;

            foreach (StringInfo stringInfo in stringInfos)
            {
                double x = stringInfo.TopPoint.X;
                double volume = stringInfo.Envelope.Volume;

                if (volume == 0)
                {
                    if (stringInfo.BezierSegment.Point1.X == x)
                    {
                        continue;
                    }
                    else
                    {
                    }
                }
                else
                {
                    if (stringInfo.BezierSegment.Point1.X == x ||
                        stringInfo.BezierSegment.Point1.X > x)
                    {
                        x -= 10 * volume;
                    }
                    else
                    {
                        x += 10 * volume;
                    }
                }

                stringInfo.BezierSegment.Point1 = new Point(x, stringInfo.BezierSegment.Point1.Y);
            }
        }
    }
}