﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Windows.Media;

namespace Petzold.MusicSynthesis
{
    public class DynamicPcmStreamSource : MediaStreamSource
    {
        const int ChannelCount = 2;
        const int BitsPerSample = 16;

        int BufferSamples;
        int BufferSize;

        MediaStreamDescription mediaStreamDescription;
        Dictionary<MediaSampleAttributeKeys, string> mediaSampleAttributes;
        MemoryStream memoryStream;
        long timestamp;

        public DynamicPcmStreamSource(int sampleRate, int audioBufferLength)
        {
            this.SampleRate = sampleRate;
            this.AudioBufferLength = audioBufferLength;
            this.BufferSamples = Math.Min(4096, audioBufferLength * sampleRate / 1000);
            this.BufferSize = ChannelCount * BufferSamples * BitsPerSample / 8;

            // Create empty mediaSampleAttributes dictionary for OpenReadAsync
            mediaSampleAttributes = new Dictionary<MediaSampleAttributeKeys, string>();

            // Create re-usable MemoryStream for accumulating audio samples
            memoryStream = new MemoryStream();
        }

        public DynamicPcmStreamSource(int sampleRate)
            : this(sampleRate, 1000)
        {
        }

        public IStereoSampleProvider SampleProvider { get; set; }

        public int SampleRate { protected set; get; }

        protected override void OpenMediaAsync()
        {
            int byteRate = SampleRate * ChannelCount * BitsPerSample / 8;
            short blockAlign = (short)(ChannelCount * (BitsPerSample / 8));

            // Build string-based wave-format structure
            string waveFormat = "";
            waveFormat += ToLittleEndianString(string.Format("{0:X4}", 1));      // indicates PCM
            waveFormat += ToLittleEndianString(string.Format("{0:X4}", ChannelCount));
            waveFormat += ToLittleEndianString(string.Format("{0:X8}", SampleRate));
            waveFormat += ToLittleEndianString(string.Format("{0:X8}", byteRate));
            waveFormat += ToLittleEndianString(string.Format("{0:X4}", blockAlign));
            waveFormat += ToLittleEndianString(string.Format("{0:X4}", BitsPerSample));
            waveFormat += ToLittleEndianString(string.Format("{0:X4}", 0));

            // Put wave format string in media streams dictionary
            var mediaStreamAttributes = new Dictionary<MediaStreamAttributeKeys, string>();
            mediaStreamAttributes[MediaStreamAttributeKeys.CodecPrivateData] = waveFormat;

            // Make description to add to available streams list
            var availableMediaStreams = new List<MediaStreamDescription>();
            mediaStreamDescription = new MediaStreamDescription(MediaStreamType.Audio, mediaStreamAttributes);
            availableMediaStreams.Add(mediaStreamDescription);

            // Set some appropriate keys in the media source dictionary
            var mediaSourceAttributes = new Dictionary<MediaSourceAttributesKeys, string>();
            mediaSourceAttributes[MediaSourceAttributesKeys.Duration] = "0";
            mediaSourceAttributes[MediaSourceAttributesKeys.CanSeek] = "false";

            // Signal that the open operation is completed
            ReportOpenMediaCompleted(mediaSourceAttributes, availableMediaStreams);
        }

        // For building string-based wave-format structure
        string ToLittleEndianString(string bigEndianString)
        {
            StringBuilder strBuilder = new StringBuilder();

            for (int i = 0; i < bigEndianString.Length; i += 2)
                strBuilder.Insert(0, bigEndianString.Substring(i, 2));

            return strBuilder.ToString();
        }

        // Provides audio samples from AudioSampleProvider property.
        //  (MediaStreamType parameter will always equal Audio.)
        protected override void GetSampleAsync(MediaStreamType mediaStreamType)
        {
            // Reset MemoryStream object
            memoryStream.Seek(0, SeekOrigin.Begin);

            // If the AudioSampleProvider is available, use it to get samples
            if (this.SampleProvider != null)
            {
                for (int sample = 0; sample < BufferSamples; sample++)
                {
                    AudioSample audioSample = SampleProvider.GetNextSample();
                    memoryStream.WriteByte((byte)(audioSample.Left & 0xFF));
                    memoryStream.WriteByte((byte)(audioSample.Left >> 8));
                    memoryStream.WriteByte((byte)(audioSample.Right & 0xFF));
                    memoryStream.WriteByte((byte)(audioSample.Right >> 8));
                }
            }
            else
            {
                // Fill up with zeros
                for (int sample = 0; sample < BufferSize; sample++)
                    memoryStream.WriteByte(0);
            }

            // Send out the sample
            ReportGetSampleCompleted(new MediaStreamSample(mediaStreamDescription,
                                                           memoryStream,
                                                           0,
                                                           BufferSize,
                                                           timestamp,
                                                           mediaSampleAttributes));
            // Prepare for next sample
            timestamp += BufferSamples * 10000000L / SampleRate;
        }

        protected override void SeekAsync(long seekToTime)
        {
            ReportSeekCompleted(seekToTime);
        }

        protected override void CloseMedia()
        {
            mediaStreamDescription = null;
        }

        // Shouldn't get a call here because only one MediaStreamDescription is supported
        protected override void SwitchMediaStreamAsync(MediaStreamDescription mediaStreamDescription)
        {
            throw new NotImplementedException();
        }

        protected override void GetDiagnosticAsync(MediaStreamSourceDiagnosticKind diagnosticKind)
        {
            throw new NotImplementedException();
        }
    }
}
