﻿using System;
using System.Windows;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Navigation;
using System.Windows.Shapes;
using Microsoft.Phone.Controls;
using Petzold.MusicSynthesis;

namespace Theremin
{
    public partial class MainPage : PhoneApplicationPage
    {
        const int SAMPLE_RATE = 22050;

        static readonly Pitch MIN_PITCH = new Pitch(Note.C, 3);
        static readonly Pitch MAX_PITCH = new Pitch(Note.C, 7);
        static readonly double MIN_FREQ = MIN_PITCH.Frequency;
        static readonly double MAX_FREQ = MAX_PITCH.Frequency;
        static readonly double MIN_FREQ_LOG2 = Math.Log(MIN_FREQ) / Math.Log(2);
        static readonly double MAX_FREQ_LOG2 = Math.Log(MAX_FREQ) / Math.Log(2);

        Mixer mixer;
        ThereminOscillator oscillator;

        double xStart;      // the X coordinate corresponding to MIN_PITCH
        int xDelta;         // the number of pixels per semitone

        int touchID = Int32.MinValue;

        public MainPage()
        {
            InitializeComponent();

            // Create Oscillator, Mixer and MixerInput objects
            oscillator = new ThereminOscillator(SAMPLE_RATE);
            mixer = new Mixer();
            mixer.Inputs.Add(new StereoToMonoMixerInput(oscillator));

            // Install event handlers
            Loaded += OnLoaded;
            Touch.FrameReported += OnTouchFrameReported;
        }

        void OnLoaded(object sender, EventArgs args)
        {
            int count = MAX_PITCH.MidiNumber - MIN_PITCH.MidiNumber;
            xDelta = (int)((ContentPanel.ActualWidth - 4) / count);
            xStart = (int)((ContentPanel.ActualWidth - count * xDelta) / 2);

            Brush cBrush = new SolidColorBrush(Colors.Red);                     // C
            Brush fBrush = new SolidColorBrush(Colors.Blue);                    // F
            Brush nBrush = this.Resources["PhoneForegroundBrush"] as Brush;     // other naturals
            Brush aBrush = new SolidColorBrush(Colors.Gray);                    // accidentals
            double x = xStart + 0.5;

            for (int i = MIN_PITCH.MidiNumber; i <= MAX_PITCH.MidiNumber; i++)
            {
                Pitch pitch = new Pitch(i);

                Line line = new Line
                {
                    X1 = x,
                    Y1 = 0,
                    X2 = x,
                    Y2 = ContentPanel.ActualHeight,
                    Stroke = pitch.Note == Note.C ? cBrush : (pitch.Note == Note.F ? fBrush : (pitch.IsNatural ? nBrush : aBrush))
                };

                ContentPanel.Children.Add(line);

                x += xDelta;
            }
        }

        protected override void  OnNavigatedTo(NavigationEventArgs args)
        {
            DynamicPcmStreamSource dynamicPcmStreamSource = new DynamicPcmStreamSource(SAMPLE_RATE, 15);
            dynamicPcmStreamSource.SampleProvider = mixer;
            mediaElement.SetSource(dynamicPcmStreamSource);

         	base.OnNavigatedTo(args);
        }

        void OnTouchFrameReported(object sender, TouchFrameEventArgs args)
        {
            TouchPointCollection touchPoints = args.GetTouchPoints(ContentPanel);

            foreach (TouchPoint touchPoint in touchPoints)
            {
                Point pt = touchPoint.Position;
                int id = touchPoint.TouchDevice.Id;

                switch (touchPoint.Action)
                {
                    case TouchAction.Down:
                        oscillator.Amplitude = ThereminOscillator.MIN_AMPLITUDE;
                        oscillator.DestinationAmplitude = CalculateAmplitude(pt.Y);
                        oscillator.Frequency = CalculateFrequency(pt.X);
                        oscillator.DestinationFrequency = oscillator.Frequency;
                        HighlightLines(pt.X, true);
                        touchID = id;
                        break;

                    case TouchAction.Move:
                        if (id == touchID)
                        {
                            oscillator.DestinationFrequency = CalculateFrequency(pt.X);
                            oscillator.DestinationAmplitude = CalculateAmplitude(pt.Y);
                            HighlightLines(pt.X, true);
                        }
                        break;

                    case TouchAction.Up:
                        if (id == touchID)
                        {
                            oscillator.DestinationAmplitude = 0;
                            touchID = Int32.MinValue;

                            // Remove highlighting
                            HighlightLines(0, false);
                        }
                        break;
                }
            }
        }

        double CalculateAmplitude(double y)
        {
            return Math.Min(1, Math.Pow(10, -4 * (1 - y / ContentPanel.ActualHeight)));
        }

        double CalculateFrequency(double x)
        {
            return Math.Pow(2, MIN_FREQ_LOG2 + (x - xStart) / xDelta / 12);
        }

        void HighlightLines(double x, bool highlight)
        {
            foreach (UIElement child in ContentPanel.Children)
                if (child is Line)
                {
                    Line line = child as Line;
                    line.StrokeThickness = highlight ? Math.Max(1, xDelta - Math.Abs(x - line.X1)) : 1;
                }
        }
    }
}